local M = {}

local fileHandler            = require("scenario/gravitationalRacing/utils/fileHandler")
local scenarioDetailsHandler = require("scenario/gravitationalRacing/scenario/scenarioDetails")
local scenarioController     = require("scenario/gravitationalRacing/scenarioController")
local celestialsHandler      = require("scenario/gravitationalRacing/celestial/celestialsHandler")
local ClassVehicle           = require("scenario/gravitationalRacing/classes/classVehicle")

local saveFileData = nil
local triggersGroup = nil
local setup = false
local sourceFile = ""

local vehicles = {player = nil}

local display = {displaying = false, scenario = ""}

local function difficultyToColour(difficulty)
  --[[
  Returns the difficulty colour (rgba)
  ]]--
  difficulty = difficulty:lower()
  if difficulty == "basic" then
    return "0 1 0 1"
  elseif difficulty == "advanced" then
    return "1 1 0 1"
  elseif difficulty == "expert" then
    return "1 0 0 1"
  elseif difficulty == "insane" then
    return "0.1171875 0.5625 1 1"
  else
    return "1 1 1 1"
  end
end

local function setupScenarios()
  --[[
  Sets up the scenarios by locking those which are not avaiable to the player,
  placing the medals achieved and indicating new scenarios which have been unlocked
  ]]--
  local scenarios = saveFileData.scenarios
  for scName, scData in pairs(scenarios) do
    local scNameFormatted = scName:gsub(" ", "_"):lower()
    local triggerName = "sc_"..scNameFormatted
    local pos = scenetree.findObject(triggerName):getPosition()

    local difficulty
    if scName ~= "Solar System Simulation" then
      difficulty = jsonDecode(readFile("levels/smallgrid/scenarios/gravitationalRacing/tracks/"..scNameFormatted..".json"))[1].name:match("%((%a+)%)")
    else
      difficulty = jsonDecode(readFile("levels/smallgrid/scenarios/gravitationalRacing/"..scNameFormatted..".json"))[1].name:match("%((%a+)%)")
    end

    local difficultyColour = difficultyToColour(difficulty)

    if not scData.unlocked then
      if not scenetree.findObject(triggerName.."_shield") then
        TorqueScript.eval([[
        new TSStatic(]]..triggerName..[[_shield) {
          shapeName = "levels/smallgrid/art/gravitationalRacing/hubWorld/scenario_shield.dae";
          position = "]]..pos.x..[[ ]]..pos.y..[[ ]]..pos.z..[[";
          scale = "12 12 12";
          collisionType = "Visible Mesh Final";
          decalType = "Visible Mesh Final";
        };
        ]])
      end
    elseif not scData.tried then
      if not scenetree.findObject(triggerName.."_waypoint") then
        TorqueScript.eval([[
        new TSStatic(]]..triggerName..[[_waypoint){
          shapeName = "art/shapes/interface/checkpoint_marker.dae";
          position = "]]..pos.x..[[ ]]..pos.y..[[ ]]..pos.z..[[";
          scale = "20 20 150";
          instanceColor = "]]..difficultyColour..[[";
        };
        ]])
      end
    end
  end
end

local function setupScenary()
  --[[
  Shows the applicable scenary, based on the scenarios completed
  ]]--
  local scenaryMap = jsonDecode(readFile("lua/scenario/gravitationalRacing/dataValues/trackRequirementMap.json"))
  for scName, data in pairs(saveFileData.scenarios) do
    --All objects are placed correctly in the prefab, so remove not achieved scenary
    if not data.completed then
      local scenary = scenaryMap[scName].hubworld
      --Not all scenarios unlock scenary
      if scenary then
        for _, objName in ipairs(scenary) do
          --Delete the object
          celestialsHandler.findCelestial(objName, nil):markForDestruction(true)
        end
      end
    end
  end
end

local function setupSideInfo()
  --[[
  Setups the side UI
  ]]--
  local data = {}
  local fileData = saveFileData.scenarios
  --Target data is in the order of scenarios
  local targetData = jsonDecode(readFile("lua/scenario/gravitationalRacing/dataValues/targetData.json"))

  local i = 1
  for k, _ in pairs(targetData) do
    local saveData = fileData[k] or error("No savedata is stored for k="..tostring(k or "nil"))
    --Only show if that sceanrio has been unlocked
    if saveData.unlocked then
      local sourceFile = "levels/smallgrid/scenarios/gravitationalRacing/tracks/"..k:gsub(" ", "_"):lower()..".json"
      local _, difficulty = scenarioDetailsHandler.getScenarioDetails(sourceFile)
      local prefabName = scenarioDetailsHandler.getSpecificDetail("prefabs", sourceFile)[1]

      --Used to order scenarios by difficulty and progression
      local stringOrder = {basic = 1, advanced = 2, expert = 3, insane = 4}

      data[i] = {
        scenarioName = k,
        scenarioColour = scenarioDetailsHandler.difficultyToColour(difficulty),
        difficulty = difficulty,
        collectable = saveData.collectables[1],
        medals = {
          resets = saveData.medals.resets,
          time = saveData.medals.time
        },
        --100 is used as some random value that will be much bigger than the total
        --number of scenarios with that difficulty
        index = stringOrder[prefabName:gsub("%d+", "")]*100 + prefabName:match("%d+")
        --The above results in index ranges:
        --  Basic    = 101-199
        --  advanced = 201-299
        --  Expert   = 301-399
        --  Insane   = 401-499
      }

      i = i + 1
    end
  end

  --Sort by calculated indexes
  table.sort(data, function(a, b) return a.index < b.index end)

  guihooks.trigger("displayStats", data)
end

--------------------------------------------------------------------------------------------------------------------
--Teleport Functions------------------------------------------------------------------------------------------------
--------------------------------------------------------------------------------------------------------------------

local teleportData = {destination = "", angle = 0, starting = false, time = 3}

local function teleportToScenario(scenario)
  --[[
  Loads the specified scenario
  ]]--
  if not scenario then
    log("E", "gravitationalRacing: teleportToScenario()", "Cannot load up a nil scenario, ignoring...")
  end

  local filePath = ""
  if scenario == "solar_system_simulation" then
    filePath = "levels/smallgrid/scenarios/gravitationalRacing/"..scenario..".json"
  else
    filePath = "levels/smallgrid/scenarios/gravitationalRacing/tracks/"..scenario..".json"
  end

  scenario_scenariosLoader.start(scenario_scenariosLoader.loadScenario(filePath, nil, filePath))
end

local function beginTeleport(dt)
  --[[
  Begins the teleport
  ]]--
  local alphas = {math.min(1, 3-teleportData.time), math.min(1, 2 - teleportData.time), math.min(1, 1 - teleportData.time)}

  for i, alpha in ipairs(alphas) do
    TorqueScript.eval('teleport'..i..'.instanceColor = "0 0 1 '..alpha..'";')
  end

  teleportData.time = teleportData.time - dt

  if teleportData.time <= 0 then
    teleportToScenario(teleportData.destination)
    --Prevents this function being called while the new scenario is being loaded
    teleportData.starting = false
  end
end

local function cancelTeleport()
  --[[
  Cancels the teleport by resetting the teleport data
  ]]--
  teleportData.destination = ""
  teleportData.starting = false
  teleportData.time = 3
  for i = 1, 3 do
    TorqueScript.eval('teleport'..i..'.position = "10000 0 0";')
    TorqueScript.eval('teleport'..i..'.instanceColor = "0 0 1 0";')
  end
end

-------------------------------------------------------------------------------------------------------------------------------------------------------

local function onPreRender(dt)
  if setup then
    celestialsHandler.update(true, vehicles, dt)
    vehicles.scenario_player0:update(dt)

    --Begin the teleporting procedure
    if teleportData.starting then
      beginTeleport(dt)
    end

    local vehPos = vehicles.scenario_player0:getPosition()

    if triggersGroup then
      if not display.displaying then
        for _, objName in ipairs(triggersGroup) do
          local objPos = scenetree.findObject(objName):getPosition()

          local dist = math.sqrt((vehPos:getX()-objPos.x)^2 + (vehPos:getY()-objPos.y)^2 + (vehPos:getZ()-objPos.z)^2)

          if dist <= 20 then
            local data = {}

            if objName ~= "sc_solar_system_simulation" then
              local name, difficulty = scenarioDetailsHandler.getScenarioDetails("levels/smallgrid/scenarios/gravitationalRacing/tracks/"..objName:gsub("sc_", "")..".json")

              local collectable
              local requirementsFormatted

              local saveData = fileHandler.readFromFile(name)
              if saveData then
                if saveData.collectables then
                  collectable = saveData.collectables[1]
                end

                local unlocked = saveData.unlocked
                if not unlocked then
                  local reqMap = jsonDecode(readFile("lua/scenario/gravitationalRacing/dataValues/trackRequirementMap.json"))
                  local currentMedals = fileHandler.getNumberOfMedals()
                  local price = reqMap[name].price - currentMedals

                  if price > 0 then
                    requirementsFormatted = " "..price
                    if price == 1 then
                      requirementsFormatted = requirementsFormatted.." medal"
                    else
                      requirementsFormatted = requirementsFormatted.." medals"
                    end
                  else
                    requirementsFormatted = ""
                  end

                  for scenario, data in pairs(reqMap) do
                    for _, sc in ipairs(data.unlocks or {}) do
                      if sc == name then
                        if requirementsFormatted ~= "" then
                          requirementsFormatted = requirementsFormatted.." or "..scenario
                        else
                          requirementsFormatted = "  "..scenario
                        end

                        --There won't be another instance of this so continue to next scenario
                        break
                      end
                    end
                  end
                end
              end

              if name == "Tutorial" then
                data = {
                  scenarioName = name,
                  difficulty = {
                    type = difficulty,
                    colour = scenarioDetailsHandler.difficultyToColour(difficulty)
                  }
                }
              else
                data = {
                  scenarioName = name,
                  difficulty = {
                    type = difficulty,
                    colour = scenarioDetailsHandler.difficultyToColour(difficulty)
                  },
                  collectable = collectable,
                  resets = {
                    best = saveData.resets,
                    wonMedal = saveData.medals.resets
                  },
                  time = {
                    best = saveData.time,
                    wonMedal = saveData.medals.time
                  },
                  requirementsFormatted = requirementsFormatted
                }
              end
            else
              data = {
                scenarioName = "Solar System Simulation",
                difficulty = {
                  type = "Simulation",
                  colour = scenarioDetailsHandler.difficultyToColour("Simulation")
                }
              }
            end

            --Update the UI
            guihooks.trigger("displayTrackInfo", data)

            --Prevent unecessary looping of this code will data is being displayed
            display.displaying = true
            display.scenario = objName
          end
        end
      else
        local objPos = scenetree.findObject(display.scenario):getPosition()
        local dist = math.sqrt((vehPos:getX()-objPos.x)^2 + (vehPos:getY()-objPos.y)^2 + (vehPos:getZ()-objPos.z)^2)
        if dist > 20 then
          --Vehicle has moved out of the area so the UI should be reset
          guihooks.trigger("hideTrackInfo")
          --Reset
          display.displaying = false
          display.scenario = ""
        end
      end
    end
  end
end

local function onBeamNGTrigger(data)
  if data.triggerName:find("sc_") then
    if data.event == "enter" then
      ui_message("Teleporting...", 3, "A")
      teleportData.destination = data.triggerName:gsub("sc_", "")
      teleportData.starting = true

      --Place markers
      local pos = scenetree.findObject(data.triggerName):getPosition()
      for i = 1, 3 do
        TorqueScript.eval('teleport'..i..'.position = "'..pos.x..' '..pos.y..' '..pos.z..'";')
      end
    else
      ui_message("Teleporting cancelled", 3, "A")
      --Reset data back to initial state
      cancelTeleport()
    end
  end
end

local function onScenarioChange(sc)
  if sc.state == "pre-running" then
    celestialsHandler.createCelestials(sourceFile ~= sc.sourceFile, true)
    sourceFile = sc.sourceFile

    --This prevents an issue where returning from a level to the hub-world and going back
    --would overwrite celestials in the scenario with hub-world celestials
    scenarioController.setSourceFile(sourceFile)

    if not vehicles.scenario_player0 then
      vehicles.scenario_player0 = ClassVehicle.new("scenario_player0", scenetree.findObject("scenario_player0"):getPosition())
    end

    --Hide track info
    guihooks.trigger("hideTrackInfo")

    saveFileData = fileHandler.readWholeFile()

    setupScenarios()
    setupScenary()
    be:reloadStaticCollision()

    triggersGroup = scenetree.findClassObjects('BeamNGTrigger')

    TorqueScript.eval([[
      teleport1.position = "10000 0 0";
      teleport2.position = "10000 0 0";
      teleport3.position = "10000 0 0";
    ]])

    setupSideInfo()

    --Hide the floor
    local ground = scenetree.findClassObjects('Groundplane')
    for _, name in ipairs(ground) do
      scenetree.findObject(name).hidden = true
    end

    setup = true

    -- require("util/richPresence").set("Definitely Didn't Mod This")
  end
end

M.onPreRender = onPreRender
M.onBeamNGTrigger = onBeamNGTrigger
M.onScenarioChange = onScenarioChange
return M
