local M = {}

local stringFormatter = require("scenario/gravitationalRacing/utils/stringFormatting")
local ClassVector     = require("scenario/gravitationalRacing/classes/classVector")
local ClassCelestial  = require("scenario/gravitationalRacing/classes/classCelestial")
local factors         = require("scenario/gravitationalRacing/dataValues/factors")
local field           = require("scenario/gravitationalRacing/celestial/gravitationalScalarField")
local tableComp       = require("scenario/gravitationalRacing/utils/tableComprehension")

local celestials = {fixedDynamic = {}, dynamic = {}, static = {}}
local initial = {positions = {length = 0}, scale = {}}
local passiveMode = false
local spectating, camera = nil, nil

local function findCelestial(name, objectType)
	--[[
	Returns the celestial object in the table of celestials in the scenario
	This function MUST be given the celestial name, but can also receieve an optional
	parameter to narrow the search (object type)
	]]--
	if objectType then
		return celestials[objectType][name]
	else
		for type, table in pairs(celestials) do
			if table[name] then
				return table[name]
			end
		end
	end

	log("E", "gravitationalRacing: findCelestial()", "Could not find celestial with name="..name..", ignoring...")
end

local function getOrbitalVelocity(celestial, parentCelestial)
	--[[
	Calculates the orbital velocity of an object, given an instance of the parent
	celestial body
	Note: Formula derived from F = GMm/r² = mv²/r, --> v = sqrt(GM/r) only works for NON-STARS
	Another formula can be derived for stars: a = F/m, F = GMm/r², a = v²/r -->  v = sqrt(GM/4r)
	Note2: The vector returned will always have the object going anti-clockwise
	around its parent celestials, and starts in the +y direction
	]]--
	local distance = parentCelestial:getPosition():subtract(celestial:getPosition())
	local r = distance:getMagnitude() / factors.getDistanceScaleFactor()

	local celestialType = celestial:getType()

	if celestialType == "star" or celestialType == "blackhole" then
		return ClassVector.new(0, math.sqrt(ClassCelestial.getGravitationalConstant() * parentCelestial:getMass()/(4*r)) * factors.getDistanceScaleFactor() * factors.getTimeScaleFactor(), 0)
	else
		return ClassVector.new(0, math.sqrt(ClassCelestial.getGravitationalConstant() * parentCelestial:getMass()/r) * factors.getDistanceScaleFactor() * factors.getTimeScaleFactor(), 0)
	end
end

local function setInitialVelocities()
	--[[
	Sets the initial velocites for all dynamic celestials
	]]--
	local setVel = function(celestial, orbittingBodyName, direction)
		--[[
		Gets and sets the orbital velocity of a celestial given its parent celestial
		]]--
		local stellarCelestial = findCelestial(orbittingBodyName)
		v = getOrbitalVelocity(celestial, stellarCelestial)

		if direction == "neg" then
			v = v:multiply(-1)
		end

		celestial:setVelocity(v)
		celestial:setupPath(stellarCelestial)
	end

	local stars = {}
  -- Set the velocity of non-stars, and store stars in another table to handle later
	for k, celestial in pairs(tableComp.merge(celestials.dynamic, celestials.fixedDynamic)) do
		if ((celestial:getType() ~= "star" and celestial:getType() ~= "blackhole") or celestial:getObjectType() == "fixedDynamic") and celestial:getParentCelestial() then
			setVel(celestial, celestial:getParentCelestial(), "pos")
		elseif celestial:getObjectType() == "fixedDynamic" then
			celestial:setupPath()
		elseif celestial:getType() == "star" then
			--Zero velocity for stars
			celestial:setVelocity(ClassVector.new(0, 0, 0))
		end
	end

	-- --Handle stars' velocities--
	-- if #stars > 2 then
	--
	-- end
	--
	--
	-- local names = {stars[1]:getName(), stars[2]:getName()}
	-- local directions = {"pos", "neg"}
	-- --To make sure they go ani-clockwise, the star who has a larger position Y
	-- --must be the one to recieve a positive acceleration in the X axis
	-- local star1InFront = stars[1]:getPosition():getY() > stars[2]:getPosition():getY()
	--
	-- for i, star in ipairs(stars) do
	-- 	name = names[#names+1-i]
	-- 	if star1InFront then
	-- 		--Sets the first ot be positive, the second to be negative
	-- 		setVel(star, name, directions[i])
	-- 	else
	-- 		--Sets the first ot be negative, the second to be positive
	-- 		setVel(star, name, directions[#directions+1-i])
	-- 	end
	-- end
end

local function setInitialConditions()
	--[[
	Sets the initial starting position and scale for each celestial
	If the information does not exist, create it and stored it
	]]--
	--Only one length variable needs to be stored for both positions and scale tables,
	--as they will have the same value
	if initial.positions.length == 0 then
		--Save initial conditions (indicates the scenario has been reloaded or first time loaded)
		for type, table in pairs(celestials) do
			initial.positions[type] = {}
			initial.scale[type] = {}
			for k, celestial in pairs(table) do
				local obj = scenetree.findObject(k)

				local objPos = obj:getPosition()
				initial.positions[type][k] = ClassVector.new(objPos.x, objPos.y, objPos.z)
				initial.positions.length = initial.positions.length + 1

				initial.scale[type][k] = obj:getScale().x
			end
		end
	end

	--Set object's position in its instance
	for type, table in pairs(celestials) do
		for k, celestial in pairs(table) do
			local pos = initial.positions[type][k]
			local scale = initial.scale[type][k]

			local initPos = ClassVector.new(pos:getX(), pos:getY(), pos:getZ())
			celestial:setPosition(initPos)
			celestial:setInitialPosition(initPos)
			celestial:setScale(scale)

			local newRadius = scale*factors.getRadiusScaleFactor()
			celestial:setRadius(newRadius)

			local newMass = celestial:calculateMass(newRadius, celestial:getActual("radius"), celestial:getActual("mass"))
			celestial:setMass(newMass)

			--Update its position and size in world
			celestial:display()
		end
	end
end

-- local function enablePassiveMode()
-- 	--[[
-- 	Deactivates the celestial's danger zones and doesn't affect the player
-- 	]]--
-- 	passiveMode = true
--
-- 	for type, groupOfCelestials in pairs(celestials) do
-- 		for k, value in pairs(groupOfCelestials) do
-- 			local instance = celestials[type][k]
-- 			instance:setPassive(true)
-- 		end
-- 	end
-- end
--
-- local function disablePassiveMode(force)
-- 	--[[
-- 	Activates the celestial's danger zones and affect the player
-- 	]]--
-- 	--Used for when the hub-world has already set passive mode and so don't override it
-- 	if passiveMode and not force then
-- 		return
-- 	end
--
-- 	passiveMode = false
--
-- 	for type, groupOfCelestials in pairs(celestials) do
-- 		for k, value in pairs(groupOfCelestials) do
-- 			local instance = celestials[type][k]
-- 			instance:setPassive(false)
-- 		end
-- 	end
-- end

local function findCelestials()
	--[[
	Finds all celestials in the scenario
	]]--
	local celestialsInScenario = {}
	local categories = ClassCelestial.getTypes()

	local i, obj
	for _, category in ipairs(categories) do
		i = 1
		obj = scenetree.findObject(category.."_"..i)
		--While they exist, add them into the table
		while obj do
			table.insert(celestialsInScenario, {category = category, obj = obj})
			i = i + 1
			obj = scenetree.findObject(category.."_"..i)
		end
	end

	return celestialsInScenario
end

local function createCelestials(fullReset, passive)
	--[[
	Creates the celestials in the scenario into objects
	]]--
	passiveMode = passive
	--When loading another scenario, the data here is still stored, so can be used for setting up
	if fullReset then
		--Reset/update variables
		celestials = {fixedDynamic = {}, dynamic = {}, static = {}}
		initial = {positions = {length = 0}, scale = {}}

		local foundCelestials = findCelestials()

		for _, celestial in ipairs(foundCelestials) do
			local objData = celestial.obj

			local name = objData.name
			local objectType = objData.objectType

			if not objectType or (objectType == "fixedDynamic" and not objData.path) then
				error("Object "..name.." has not been properly setup as a celestial!")
			end

			local celestialType = name:gsub("_%d+", "")
			celestials[objectType][name] = ClassCelestial.new(name, celestialType, objData:getScale().x, objectType, objData.orbittingBody, passiveMode, objData.delayed or false, objData.axisTilt)
		end
	else
		--Setup danger zone areas and delete old trails
		for type, groupOfCelestials in pairs(celestials) do
			for k, value in pairs(groupOfCelestials) do
				local instance = celestials[type][k]

				if not passiveMode then
					instance:displayRadialZones()
				end

				instance:reset()
			end
		end
	end

	setInitialConditions()
	--Note: velocities need to be set after as to calculate this the positions
	--need to be setup beforehand
	setInitialVelocities()
end

local function update(start, vehicles, dt)
	--[[
	Updates all celestials in the scenario
	]]--
	if bullettime.get() > 0 then
		local attactPlayer = function(celestial, invert)
			--[[
			Attracts the player car to the celestial
			]]--
			local forceOnPlayer = celestial:getForce(vehicles.scenario_player0)
			if invert then
				forceOnPlayer = forceOnPlayer:multiply(-1)
			end

			vehicles.scenario_player0:addWind(forceOnPlayer)
			celestial:handleCollision(vehicles.scenario_player0)
		end

		local actualDt = dt * bullettime.get()

		--Dynamic celestials wait until the scenario start
		if start then
			for k, celestial1 in pairs(celestials.dynamic) do
				if not celestial1:isRemoved() then
					--Account for the possibility that the object has been destroyed previously
					--by another object
					if not celestial1:isMarkedForDestruction() then
						for k2, celestial2 in pairs(celestials.dynamic) do
							--Do not attract object to itself
							if k ~= k2 then
								if not celestial2:isMarkedForDestruction() and not celestial2:isRemoved() then
									local force = celestial1:getForce(celestial2)
									celestial1:applyForce(force, actualDt)
									celestial1:handleCollision(celestial2)
								end
							end
						end
					end

					if not passiveMode then
						--Affect the player
						attactPlayer(celestial1)
					end

					celestial1:update(actualDt)
				end
			end
		end

		for k, celestial in pairs(celestials.fixedDynamic) do
			if not celestial:isRemoved() then
				if start and not passiveMode then
					attactPlayer(celestial, true)
				end

				celestial:update(actualDt)
			end
		end

		for k, celestial in pairs(celestials.static) do
			if not celestial:isRemoved() then
				if start and not passiveMode then
					attactPlayer(celestial, true)
				end

				celestial:update(actualDt)
			end
		end

		if start then
			vehicles.scenario_player0:update(actualDt)
		end

		if spectating then
			local pos = spectating:getPosition()
			local radius = spectating:getScaledRadius()
			camera:setPosition(vec3(pos:getX(), pos:getY(), pos:getZ() + 250):toPoint3F())
		end
	end
end

local function createCelestial(shapeName, initPos, name, type, scale, objectType, orbittingBody, passive, delayed, rotation)
	--[[
	Creates a new celestial
	]]--
	--Create the object
	if not scenetree.findObject(name) then
		TorqueScript.eval([[
		new TSStatic(]]..(name)..[[) {
			shapeName = "levels/smallgrid/art/gravitationalRacing/celestialbodies/]]..shapeName..[[.dae";
			dynamic = "1";
			scale = "]]..scale..[[ ]]..scale..[[ ]]..scale..[[";
			objectType = "dynamic";
		};
		]])
	end

	--Create the associated instance
	local obj = ClassCelestial.new(name, type, scale, objectType, orbittingBody, passive, delayed, rotation)
	obj:setPosition(initPos)
	--Add it into the table for updating
	celestials[objectType][name] = obj
end

local function deleteCelestial(name, objectType)
	--[[
	Deletes a celestial
	]]--
	if objectType then
		for k, instance in pairs(celestials[objectType]) do
			if k == name then
				instance:delete()
				celestials[objectType][k] = nil
				print("Deleted "..name)
			end
		end
	else
		for i, group in ipairs(celestials) do
			for k, instance in pairs(group) do
				if k == name then
					instance:delete()
					celestials[i][k] = nil
				end
			end
		end
	end
end

local function setSpectate(name, objectType)
	--[[
	Spectates an object
	NOTE: Experimental
	]]--
	spectating = findCelestial(name, objectType)
	commands.setFreeCamera()
	camera = commands.getCamera(commands.getGame())
end

local function createForceField()
	--[[
	Creates a force scale field, showing the strength of the gravitational force
	over an area
	]]--
	field.create(celestials)
end

local function printResults()
	--[[
	A testing function for printing the various celestial related data it has created
	]]--
	print("Celestials: [Passive Mode = "..tostring(passiveMode).."]")
	print()
	local celestialInfo = {}
	local keys = ClassCelestial.toListKeys()

	for k, t in pairs(celestials) do
		for k2, instance in pairs(t) do
			table.insert(celestialInfo, instance:toList())
		end
	end

	stringFormatter.printClassInfo(celestialInfo, keys)

	print()
	print("Scenario running with factors:")
	print("  Distance Factor: "..factors.getDistanceScaleFactor())
	print("  Time Factor: "..factors.getTimeScaleFactor())
	print("  Radius Factor: "..factors.getActualRadiusScaleFactor())
	print()
end

M.findCelestial = findCelestial
-- M.enablePassiveMode = enablePassiveMode
-- M.disablePassiveMode = disablePassiveMode
M.createCelestials = createCelestials
M.update = update
M.createForceField = createForceField
M.createCelestial = createCelestial
M.deleteCelestial = deleteCelestial
M.setSpectate = setSpectate
M.printResults = printResults
return M
