ClassCelestial = {}
ClassCelestial.__index = ClassCelestial

local factors       = require("scenario/gravitationalRacing/dataValues/factors")
local ClassText     = require("scenario/gravitationalRacing/classes/classText")
local ClassPath     = require("scenario/gravitationalRacing/classes/classPath")
local ClassVector   = require("scenario/gravitationalRacing/classes/classVector")
local ClassMatrix   = require("scenario/gravitationalRacing/classes/classMatrix")
local field         = require("scenario/gravitationalRacing/celestial/gravitationalScalarField")
local celestialInfo = require("scenario/gravitationalRacing/dataValues/celestialInformation")

--Instance methods
local function getGravitationalConstant()
  --[[
  Returns the value of G
  ]]--
  return 6.67408*math.pow(10, -11)
end

function ClassCelestial:calculateMass(radius, actualRadius, actualMass)
  --[[
  Returns the mass of this object
  ]]--
  if not radius or not actualRadius then
    error("One or more parameters is nil")
  end

  --Given m = ρV, it follows that m ∝ r³:
  --thus, the mass is affected by the ratio of (radius / actualRadius)³
  local radiusRatio = math.pow(radius / actualRadius, 3)

  return actualMass * radiusRatio
end

function ClassCelestial:calculateRotationSpeed(period)
  --[[
  Returns the rotation speed of this object
  ]]--
  local freq = 1/period
  return 2*math.pi*freq
end

function ClassCelestial:new(name, type, scale, objectType, orbittingBody, passive, delayed, axisTilt)
  --Some parameters are optional as not required/applicable in some cases
  if not name or not type or not scale or not objectType then
    error("One or more parameters is nil! [name="..tostring(name or "nil")..", type="..tostring(type or "nil")..", scale="..tostring(scale or "nil")..", objectType="..tostring(objectType or "nil").."]")
  end

  local self = {}
  self.name = name
  self.shapeName = scenetree.findObject(name).shapeName
  --Whether this object is fixed in position (static), manipulable (dynamic) or fixed along a path (fixedDynamic)
  self.objectType = objectType
  --The celestial type
  self.type = type
  self.delayed = delayed

  self.orbittingBody = orbittingBody
  self.isPassive = passive

  self.position        = ClassVector.new(0, 0, 0)
  self.initialPosition = ClassVector.new(0, 0, 0)
  self.acceleration    = ClassVector.new(0, 0, 0)
  self.velocity        = ClassVector.new(0, 0, 0)

  --Since the framerate will change, this is used to scale the velocity appropriately
  self.frameTime = 1

  self.trail = {}
  self.trailLimit = 100
  self.path = nil
  self.placedPath = false

  self.label = nil

  self.shouldBeDestroyed = false
  --Only used when the object should be deleted no matter what
  self.forceDelete = false
  self.isDestroyed = false

  self.radius = scale*factors.getRadiusScaleFactor()

  local celestial = self.shapeName:gsub("levels/smallgrid/art/gravitationalRacing/celestialbodies/", ""):gsub(".dae", "")
  --For blackholes
  celestial = celestial:gsub("_eventHorizon", "")

  self.info = celestialInfo.getData(celestial)

  self.mass = ClassCelestial:calculateMass(self.radius, self.info.radius, self.info.mass)

  ---Rotation can be specified for axis tilt instead of real-life value
  self.axisTilt = axisTilt or self.info.axisTilt
  self.rotAngle = 0
  self.angularVelocity = ClassCelestial:calculateRotationSpeed(self.info.rotationPeriod)

  self.setup = false

  setmetatable(self, ClassCelestial)
  return self
end

function ClassCelestial:getActual(attribute)
  return self.info[attribute] or log("E", "celestialInformaton.getData()", "Cannot get an unknown attribute of this celestial! [attribute="..tostring(attribute).."]")
end

function ClassCelestial:createLabel()
  --[[
  Creates a 3d label above the celestial
  ]]--
  --Do not recreate an existing label
  if not self.label then
    local pos = self.position
    self.label = ClassText.new(scenetree.findObject(self.name).shapeName:gsub("levels/smallgrid/art/gravitationalRacing/celestialbodies/", ""):gsub(".dae", ""), ClassVector.new(pos:getX(), pos:getY() + self:getScaledRadius() + 5, pos:getZ()))
  end
end

function ClassCelestial:isMarkedForDestruction()
  return self.shouldBeDestroyed
end

function ClassCelestial:markForDestruction(force)
  if force then
    --Forces the object to be delete (bypasses "static" rule in update())
    self.forceDelete = true
  else
    self.shouldBeDestroyed = true
  end
end

function ClassCelestial:isDelayed()
  return self.delayed
end

function ClassCelestial:reset()
  self:destroyTrail()
  self:resetPath()
  self.acceleration:zero()
  self.velocity:zero()
  self.frameTime = 1
  self.isDestroyed = false
end

function ClassCelestial:isRemoved()
  return self.isDestroyed
end

function ClassCelestial:destroyTrail()
  --[[
  Destroys the object's trail
  ]]--
  if self.objectType == "dynamic" then
    for i = 1, #self.trail do
      if scenetree.findObject(self.name..'_trail_'..i) then
        TorqueScript.eval(self.name..'_trail_'..i..'.delete();')
      end
    end

    self.trail = {}
  elseif self.objectType == "fixedDynamic" then
    for i = 1, #self.trail do
      if scenetree.findObject(self.name..'_trail_'..i) then
        TorqueScript.eval(self.name..'_trail_'..i..'.hidden = "true";')
      end
    end
  end

end

function ClassCelestial:delete()
  --[[
  Destroys this instance
  ]]--
  self:removeFromScene()
  self = nil
end

function ClassCelestial:removeFromScene()
  --[[
  Removes the celestial from view and prevents it doing anything
  Note: The object is not deleted from the scenario, only positioned away from view,
  and reduced to 0 scale so it cannot be seen or have any effect
  This is because it is computationally expensive to delete and create a
  numerous number of objects
  This is NOT the same as delete() as deleting destroys the entire instance
  ]]--
  TorqueScript.eval(self.name..'.position = "10000 -10000 10000";')
  TorqueScript.eval(self.name..'.scale = "0 0 0";')

  if self.type == "blackhole" then
    TorqueScript.eval(self.name..'_disc.position = "100000 -100000 100000";')
    TorqueScript.eval(self.name..'_disc.scale = "0 0 0";')
  end

  print(self.name.." has been destroyed!")

  self:destroyTrail()

  if self.label then
    self.label:move(ClassVector.new(10000, -10000, 10000))
  end

  self.isDestroyed = true
end

--Functions only applicable to celestials of objectType="fixedDynamic"
function ClassCelestial:setupPath(parentCelestial)
  --[[
  Setups the path (if it is fixed)
  ]]--
  --Only setup path if it has not been already
  if not self.path then
    local path = scenetree.findObject(self.name).path
    if path then
      --Note: not all of these need to be specified for certain path types
      local invertDirection = tonumber(scenetree.findObject(self.name).invertDirection)
      local frequency = tonumber(scenetree.findObject(self.name).frequency)
      local direction = scenetree.findObject(self.name).direction
      local length = tonumber(scenetree.findObject(self.name).length)

      self.path = ClassPath.newPath(path, invertDirection, self, parentCelestial, frequency, direction, length)

      self:placeTrailPath(path, parentCelestial)
    end
  end
end

function ClassCelestial:placeTrailPath(path, parentCelestial)
  --[[
  Places a trail for the celestial
  ]]--
  if path == "circular" then
    local isBinary = (self.type == "star" and parentCelestial:getType() == "star") or (self.type == "blackhole" and parentCelestial:getType() == "blackhole")

    if not (isBinary and parentCelestial:hasPlacedPath()) then
      local parentPos = parentCelestial:getPosition()

      local radius = self.position:subtract(parentPos):getMagnitude()
      if isBinary then
        --Again, assumes the masses are the same
        radius = radius/2
      end

      local circumference = 2 * math.pi * radius
      --The number of trails should not exceed the trail limit
      local numTrails = math.min(math.floor(circumference / 10), self.trailLimit)
      local angleIncrement = 2*math.pi/numTrails

      for theta = 0, 2*math.pi, angleIncrement do
        if isBinary then
          table.insert(self.trail, ClassVector.new(parentPos:getX() + radius + radius*math.cos(theta), parentPos:getY() + radius*math.sin(theta), self.position:getZ()))
        else
          table.insert(self.trail, ClassVector.new(parentPos:getX() + radius*math.cos(theta), parentPos:getY() + radius*math.sin(theta), self.position:getZ()))
        end
      end
    end
  elseif path == "linear" then
    local bounds = self.path:getBounds()
    local direction = self.path:getDirection()

    --There should be at least 5 meters between each cube
    local increment = math.max(5, math.abs(bounds[1]-bounds[2]) / self.trailLimit)

    --Loop from biggest to smallest
    for i = math.min(bounds[1], bounds[2]), math.max(bounds[1], bounds[2]), increment do
      if direction == "X" then
        table.insert(self.trail, ClassVector.new(i, self.position:getY(), self.position:getZ()))
      elseif direction == "Y" then
        table.insert(self.trail, ClassVector.new(self.position:getX(), i, self.position:getZ()))
      end
    end
  end

  self:placeTrail(true)
  self.placedPath = true
end

function ClassCelestial:resetPath()
  --[[
  Resets the path
  ]]--
  for i = 1, #self.trail do
    if scenetree.findObject(self.name..'_trail_'..i) then
      TorqueScript.eval(self.name..'_trail_'..i..'.hidden = "false";')
    end
  end

  if self.path then
    self.path:reset()
  end
end

function ClassCelestial:getPath()
  return self.path
end

function ClassCelestial:hasPlacedPath()
  return self.placedPath
end
--

function ClassCelestial:setPassive(passive)
  self.isPassive = passive
end

function ClassCelestial:setPosition(newPosition)
  --[[
  Sets the velocity of this object
  ]]--
  self.position:setX(newPosition:getX())
  self.position:setY(newPosition:getY())
  self.position:setZ(newPosition:getZ())

  --Adjust label to new position
  if self.label then
    self.label:move(ClassVector.new(newPosition:getX(), newPosition:getY() + self:getScaledRadius() + 5, newPosition:getZ()))
  end
end

function ClassCelestial:setInitialPosition(pos)
  --[[
  Sets the velocity of this object
  ]]--
  self.initialPosition:setX(pos:getX())
  self.initialPosition:setY(pos:getY())
  self.initialPosition:setZ(pos:getZ())
end

function ClassCelestial:setVelocity(newVelocity)
  --[[
  Sets the velocity of this object
  ]]--
  self.velocity:setX(newVelocity:getX())
  self.velocity:setY(newVelocity:getY())
  self.velocity:setZ(newVelocity:getZ())
end

function ClassCelestial:placeTrail(placeAll)
  --[[
  Creates and places the trail at this object's position
  ]]--
  local create = function(i)
    --[[
    Creates the trail object, if it does not exist
    ]]--
    --If it doesn't exists in the level, then create it
    if not scenetree.findObject(self.name..'_trail_'..i) then
      TorqueScript.eval([[
      new TSStatic(]]..(self.name..'_trail_'..i)..[[) {
        shapeName = "levels/smallgrid/art/gravitationalRacing/trail.dae";
        scale = "2 2 2";
        dynamic = "1";
      };
      ]])
    end
  end

  if placeAll then
    --Loop through current list of trails
    for i = 1, #self.trail do
      create(i)

      local trailPos = self.trail[i]
      TorqueScript.eval(self.name..'_trail_'..i..'.position = "'..trailPos:getX()..' '..trailPos:getY()..' '..trailPos:getZ()..'";')
    end
    -- else
    --   index = index or #self.trail
    --   if index > 0 then
    --     create(index)
    --     local trailPos = self.trail[index]
    --     if self.name=="planet_1" then print(index, trailPos:toString()) end
    --     TorqueScript.eval(self.name..'_trail_'..trailNum..'.position = "'..trailPos:getX()..' '..trailPos:getY()..' '..trailPos:getZ()..'";')
    --   end
  end
end

function ClassCelestial:getVelocity()
  return self.velocity
end

function ClassCelestial:getAngularVelocity()
  return self.angularVelocity
end

function ClassCelestial:getScale()
  return scenetree.findObject(self.name):getScale().x
end

function ClassCelestial:setAngularVelocityRelative(percentage, value)
  --[[
  Sets the rotational speed relative to what it was previously
  ]]--
  --f ∝ 1/T
  if percentage then
    self.angularVelocity = self.angularVelocity * 1/value
  else
    self.angularVelocity = self.angularVelocity + value
  end
end

function ClassCelestial:setScale(scale)
  TorqueScript.eval(self.name..'.scale = "'..scale..' '..scale..' '..scale..'";')
end

function ClassCelestial:setScaleRelative(scale)
  --[[
  Sets the scale relative to what it was previously
  Ie. a scale of 2x is double the size
  ]]--
  if scale < 0 then
    log("E", "gravitationalRacing: setScaleRelative()", "Scale cannot be less than 0, ignoring...")
    return
  end

  local previousScale = self:getScale()
  local newScale = scale * previousScale

  if newScale < 0 then
    self.shouldBeDestroyed = true
  else
    self:setScale(newScale)
  end
end

function ClassCelestial:setScale(scale)
  --[[
  Set the absolute scale of the object
  ]]--
  if scale < 0 then
    log("E", "gravitationalRacing: setScale()", "Scale cannot be less than 0, ignoring...")
    return
  end

  TorqueScript.eval(self.name..'.scale = "'..scale..' '..scale..' '..scale..'";')
end

function ClassCelestial:setMass(mass)
  --[[
  Sets the mass
  Note: setting the mass to <= 0 destroys the object
  ]]--
  if mass > 0 then
    self.mass = mass
  else
    self.mass = 0
    self.shouldBeDestroyed = true
  end
end

function ClassCelestial:setMassRelative(scale)
  --[[
  Sets the mass relative to what it was previously
  Ie. a mass of 2x is double the mass
  ]]--
  self:setMass(scale * self.mass)
end

function ClassCelestial:getMass()
  return self.mass
end

function ClassCelestial:getRadius()
  return self.radius
end

function ClassCelestial:getScaledRadius()
  --[[
  Returns the radius in the scenario
  ]]--
  --A scale of 1 means the radius of the object is 0.03m
  return self:getScale() * 0.03
end

function ClassCelestial:setRadius(radius)
  --[[
  Sets the radius
  Note: setting the radius to <= 0 destroys the object
  ]]--
  if radius > 0 then
    self.radius = radius
  else
    self.radius = 0
    self.shouldBeDestroyed = true
  end
end

function ClassCelestial:setRadiusRelative(scale)
  --[[
  Sets the radius relative to what it was previously
  Ie. a mass of 2x is double the mass
  ]]--
  self:setRadius(scale * self.radius)
end

function ClassCelestial:getPosition()
  return self.position
end

function ClassCelestial:getParentCelestial()
  return self.orbittingBody
end

function ClassCelestial:getName()
  return self.name
end

function ClassCelestial:getShapeName()
  return self.shapeName
end


function ClassCelestial:getType()
  return self.type
end

function ClassCelestial:getObjectType()
  return self.objectType
end

function ClassCelestial:instanceOf()
  return "ClassCelestial"
end

function ClassCelestial:collide(object)
  --[[
  Handles the collision of two objects: the one with the bigger mass gains what
  the other looses
  ]]--
  if not object then
    log("E", "Solar System: collide()", "Cannot collide with nil object, ignoring...")
    return
  end

  --Determine how much the objects are intersecting one another by; this distance
  --will be what is to be shifted around between objects
  local intersection = (object:getScaledRadius() + self:getScaledRadius()) - object:getPosition():subtract(self.position):getMagnitude()

  local objectToLose, objectToGain
  --Give preferential treatment to this object if they both have the same amss
  if self.mass >= object:getMass() then
    objectToLose = object
    objectToGain = self
  else
    objectToLose = self
    objectToGain = object
  end

  --The amount (0 < scale < 1) the object is clipping by
  --This is the amount to remove from the radius, and thus the size and mass
  local clippingAmount = intersection / objectToLose:getScaledRadius()
  local newScale = 1 - clippingAmount

  if newScale < 0 then
    --Object is entirely inside the other, thus should be destroyed
    objectToLose:markForDestruction()
  else
    local update = function(object, scale)
      --[[
      Updates the object's new mass, scale, radius and rotation speed
      ]]--
      --m ∝ r³, therefore if the radius doubles then the mass octuples
      local massScale = scale^3
      local radiusScale = scale

      object:setScaleRelative(scale)
      object:setMassRelative(massScale)
      object:setRadiusRelative(radiusScale)

      -- --T ∝ m, T ∝ r²
      -- object:setAngularVelocityRelative(true, 1 + massScale * radiusScale^2)
    end

    local clippingAmount2 = intersection / objectToGain:getScaledRadius()
    local newScale2 = 1 + clippingAmount2

    -- local angVelOfChunk = objectToLose:getAngularVelocity() * (clippingAmount^3 + clippingAmount^2)

    update(objectToLose, newScale)
    update(objectToGain, newScale2)

    -- --Also add the angular velocity of the additional piece to the winning object
    -- objectToGain:setAngularVelocityRelative(false, angVelOfChunk)
  end
end

function ClassCelestial:handleCollision(object)
  --[[
  Handles the collision of two objects, if necessary
  ]]--
  if not object then
    log("E", "Solar System: handleCollision()", "Cannot check for collision with nil object, ignoring...")
    return
  end

  --Distance between this object and the other
  local objPos = object:getPosition()
  local distance = objPos:subtract(self.position)

  local r = distance:getMagnitude()

  if object:instanceOf() == "ClassCelestial" and r < object:getScaledRadius() + self:getScaledRadius() then
    self:collide(object)
    --If a vehicle is inside a celestial (hence must not be static)
  elseif self.objectType ~= "static" and object:instanceOf() == "ClassVehicle" and r < self:getScaledRadius()and not object:isResetting() then
    object:scheduleReset()
    if self.type == "star" then
      object:ignite()
    end
  end
end

function ClassCelestial:findRadialZones()
  --[[
  Returns the zone radius of the gravitational field, with increasing acceleration
  ]]--
  local findRadiusWithAccel = function(accel)
    return math.sqrt(50000 * getGravitationalConstant() * 1 * self.mass / accel) * factors.getDistanceScaleFactor()
  end

  local maxAccels = {51, 101, 151}
  local rmin, rmax = findRadiusWithAccel(maxAccels[1]), findRadiusWithAccel(maxAccels[3])

  -- print("Middle: "..(50000*getGravitationalConstant()*self.mass / (((rmax-rmin)/2 + rmin)/factors.getDistanceScaleFactor())^2), rmin, rmax)

  return {[tostring(maxAccels[1])] = rmin, [tostring(maxAccels[2])] = (rmax-rmin)/2 + rmin, [tostring(maxAccels[3])] = rmax}
end

function ClassCelestial:displayRadialZones()
  --[[
  Displays a sphere, representing the point of no return for a vehicle
  ]]--
  local radii = self:findRadialZones()

  for k, v in pairs(radii) do
    local objectDangerRadiusName = self.name.."_Danger_radius"..k

    local colour = field.displayColour(tonumber(k))

    if not scenetree.findObject(objectDangerRadiusName) then
      TorqueScript.eval([[
      new TSStatic(]]..objectDangerRadiusName..[[) {
        shapeName = "levels/smallgrid/art/gravitationalRacing/field/spheres/]]..colour..[[.dae";
        scale = "]]..v..[[ ]]..v..[[ ]]..v..[[";
        dynamic = "1";
      };
      ]])
    end

    TorqueScript.eval(objectDangerRadiusName..'.position = "'..self.position:getX()..' '..self.position:getY()..' '..self.position:getZ()..'";')
  end
end

function ClassCelestial:getForceAtPoint(point)
  --[[
  Returns the force exerted at a point by this object (with mass m=1)
  ]]--
  --Distance between this object and the other
  local objPos = object:getPosition()
  local distanceVec = objPos:subtract(self.position)
  local distance = distanceVec:getMagnitude()

  --Newton's Law of Gravitation: F = GMm/r²
  local force = getGravitationalConstant()*self.mass*object:getMass() / math.pow(distance, 2)

  --Component magnitudes = current * ratio between new and old magnitudes
  return distanceVec:multiply(force/distance)
end

function ClassCelestial:getForce(object)
  --[[
  Returns the force exerted on each object by their gravity
  ]]--
  if not object then
    log("E", "Solar System: getForce()", "Cannot determine gravitational force with nil object, ignoring...")
    return
  end


  if object:instanceOf() == "ClassVehicle" and object:isImmune() then
    return ClassVector.new(0, 0, 0)
  end

  --Distance between this object and the other
  local objPos = object:getPosition()
  local distanceVec = objPos:subtract(self.position)
  local distance = distanceVec:getMagnitude() / factors.getDistanceScaleFactor()

  --Newton's Law of Gravitation: F = GMm/r²
  local force = getGravitationalConstant()*self.mass*object:getMass() / math.pow(distance, 2)

  --Component magnitudes = current * ratio between new and old magnitudes
  return distanceVec:multiply(force/distance)
end

function ClassCelestial:applyForce(force, dt)
  --[[
  Applies the force to this object
  ]]--
  if not force then
    log("E", "Solar System: applyForce()", "Cannot apply a nil force, ignoring...")
    return
  end

  --Newton's second law: F = ma --> a = F/m
  local a = force:divide(self.mass)

  --a ∝ v² ∝ t⁻² --> acts as t² as we are speeding up time (therefore decreasing time overall)
  a = a:multiply((factors.getTimeScaleFactor()*dt)^2)

  self.acceleration = self.acceleration:add(a)
end

function ClassCelestial:display()
  --[[
  Updates the visuals of this object, such as moving it in-world and placing trails
  ]]--
  if self.objectType == "dynamic" or self.objectType == "fixedDynamic" or not self.setup then
    --Do not display radial zones if this object is not setup, as passive mode
    --does not kick in until the objects are setup
    if not self.isPassive then
      self:displayRadialZones()
    end

    --Place accretion disc and jets for blackhole
    if self.name:find("blackhole") then
      TorqueScript.eval(self.name..'_disc.position = "'..self.position:getX()..' '..self.position:getY()..' '..self.position:getZ()..'";')
    end

    --Place itself
    TorqueScript.eval(self.name..'.position = "'..self.position:getX()..' '..self.position:getY()..' '..self.position:getZ()..'";')

    if not self.setup then self.setup = true end
  end
end

function ClassCelestial:updateTrail()
  --[[
  Updates the trail by adding/removing trail points as necessary
  ]]--
  --Only add in the trail if the celestial has moved at least 10 meters
  if #self.trail == 0 or self.trail[#self.trail]:subtract(self.position):getMagnitude() >= 10 then
    table.insert(self.trail, ClassVector.new(self.position:getX(), self.position:getY(), self.position:getZ()))

    if #self.trail > self.trailLimit then
      --Delete the oldest value
      table.remove(self.trail, 1)
    end

    self:placeTrail(true)
  end
end

function ClassCelestial:rotate(dt)
  --[[
  Rotates the object
  ]]--
  --Black holes do not rotate (currently)
  if self.type ~= "blackhole" then
    self.rotAngle = self.rotAngle + dt*self.angularVelocity

    local z = self.rotAngle
    local x = math.rad(self.axisTilt)

    local q = quatFromEuler(x, 0, z):toTorqueQuat()

    TorqueScript.eval(self.name..'.rotation = "'..q.x..' '..q.y..' '..q.z..' '..q.w..'";')
  end
end

function ClassCelestial:update(dt)
  --[[
  Updates the celestial for the next frame
  Note: this function returns false if it could not update, due to this object
  being destroyed
  ]]--
  --Static objects cannot be destroyed, even if they have 0 mass/radius, unless forced to
  if self.forceDelete or (self.shouldBeDestroyed and self.objectType ~= "static") then
    self:removeFromScene()
    return false
  end

  if self.objectType == "dynamic" then
    --Update the velocity from the acceleration
    self.velocity = self.velocity:add(self.acceleration)

    --The framerate will fluctuate, so the velocity should scale appropriately
    self.velocity = self.velocity:multiply(dt/self.frameTime)

    --Update the position with respect to the velocity
    self.position = self.position:add(self.velocity)

    self:updateTrail()

    self.acceleration:zero()

    self.frameTime = dt
  elseif self.objectType == "fixedDynamic" then
    self.position = self.path:update(dt)
  end

  if self.label then
    local pos = self.position:clone()
    self.label:move(ClassVector.new(pos:getX(), pos:getY() + self:getScaledRadius() + 5, pos:getZ()))
  end

  self:rotate(dt)
  self:display()

  return true
end

function ClassCelestial:toList()
  --[[
  Returns a list representation of this instance
  ]]--
  return {"ClassCelestial", self.name, scenetree.findObject(self.name).shapeName:gsub("levels/smallgrid/art/gravitationalRacing/celestialbodies/", ""):gsub(".dae", ""), self.mass, self.objectType, self:findRadialZones()["151"], self.position:toString(), self.acceleration:toString(), self.velocity:toString()}
end

function ClassCelestial:toString()
  --[[
  Returns a string representation of this instance
  ]]--
  return "ClassCelestial[Name: "..self.name..
  ", Type: "..self.type..
  ", Model: "..scenetree.findObject(self.name).shapeName:gsub("levels/smallgrid/art/gravitationalRacing/celestialbodies/", ""):gsub(".dae", "")..
  ", Mass: "..self.mass..
  ", ObjectType: "..self.objectType..
  ", DangerRadius: "..self:findRadialZones()["151"]..
  ", Position: "..self.position:toString()..
  ", Velocity: "..self.velocity:toString().."]"
end

------------------------------------------------------------------------------------------------------------------------------------------

local function new(name, type, scale, objectType, orbittingBody, passive, delayed, rotation)
  return ClassCelestial:new(name, type, scale, objectType, orbittingBody, passive, delayed, rotation)
end

local function getTypes()
  --[[
  Returns the types of celestials possible
  ]]--
  --DO NOT change the order of these: "star" must be first for some algorithms to work
  return {"star", "planet", "blackhole"}
end

local function toListKeys()
  --[[
  Returns a list with the keys of attributes in this class returned by toList()
  Note: this function exists instead of having these keys in the same table as toList()
  because order needs to be maintained, which string keys does not provide
  ]]--
  return {"Class", "Name", "Model", "Mass", "ObjectType", "DangerRadius", "Position", "Acceleration", "Velocity"}
end

return {new = new, getTypes = getTypes, toListKeys = toListKeys, getGravitationalConstant = getGravitationalConstant}
