ClassPath = {}
ClassPath.__index = ClassPath

local ClassVector = require("scenario/gravitationalRacing/classes/classVector")
local factors = require("scenario/gravitationalRacing/dataValues/factors")

function ClassPath:new(attributes)
  if not attributes then
    error("One or more parameters is nil")
  end

  local self = {}
  for k, v in pairs(attributes) do
    self[k] = v
  end

  setmetatable(self, ClassPath)
  return self
end

----------------------------------------------------------------------------------------------------------------------------------------------

ClassCircularPath = {}
ClassCircularPath.__index = ClassCircularPath

function ClassCircularPath:calculateOrbit(celestial, parentCelestial, frequency)
  --[[
  Calculates and returns the speed and raidus of the orbital path
  The optional parameter frequency is used to set a specific frequnecy, rather than calculate it
  ]]--
  local distance = parentCelestial:getPosition():subtract(celestial:getPosition())
  local radius = distance:getMagnitude()

  if not frequency then
    frequency = celestial:getVelocity():getMagnitude()/(2*math.pi*radius)
  end

  return radius, frequency
end

function ClassCircularPath:new(celestial, parentCelestial, invertDirection, frequency)
  if not celestial or not parentCelestial then
    error("One or more parameters is nil!")
  end

  local radius, frequency = ClassCircularPath:calculateOrbit(celestial, parentCelestial, frequency)

  local parentPos, selfPos = parentCelestial:getPosition(), celestial:getPosition()
  local parentType, selfType = celestial:getType(), parentCelestial:getType()

  local offset, angle
  if (selfType == "star" and parentType == "star") or (selfType == "blackhole" and parentType == "blackhole") then
    radius = radius/2
    local radiusVec = selfPos:subtract(parentCelestial:getPosition()):divide(2)
    --One star must be at a phase difference of 180
    if not parentCelestial:getPath() then
      angle = math.pi
      offset = selfPos:subtract(radiusVec)
    else
      offset = selfPos:subtract(radiusVec)
    end
  else
    offset = ClassVector.new(parentPos:getX(), parentPos:getY(), selfPos:getZ())
  end

  angle = angle and angle or 0

  local self = ClassPath:new({
    type       = "circular",
    radius     = radius,
    frequency  = frequency,
    angle      = angle,
    angleReset = angle,
    offset     = offset,
    direction  = (invertDirection and invertDirection == 1 and -1) or 1
  }
)

setmetatable(self, ClassCircularPath)
return self
end

--Inheritance
setmetatable(ClassCircularPath, {__index = ClassPath})

function ClassCircularPath:update(dt)
  --[[
  Finds the next position this object will be in, and returns it
  ]]--
  self.angle = self.angle + (dt*self.frequency) * 2*math.pi * self.direction
  return ClassVector.new(self.offset:getX() + self.radius*math.cos(self.angle), self.offset:getY() + self.radius*math.sin(self.angle), self.offset:getZ())
end

function ClassCircularPath:reset()
  --[[
  Resets the path to the beginning
  ]]--
  self.angle = self.angleReset
end

-----------------------------------------------------------------------------------------------------------------------------------------------

ClassLinearPath = {}
ClassLinearPath.__index = ClassLinearPath

function ClassLinearPath:new(celestial, invertDirection, frequency, length, direction)
  if not celestial or not frequency or not length or not direction then
    error("One or more parameters is nil! [celestial="..tostring(celestial and celestial or 'nil')..", frequency="..tostring(frequency and frequency or 'nil')..", length="..tostring(length and length or 'nil')..", direction="..tostring(direction and direction or 'nil').."]")
  end

  local celestialPos = celestial:getPosition()
  local lengthVec

  if direction == "X" then
    lengthVec = ClassVector.new(length, 0, 0)
  elseif direction == "Y" then
    lengthVec = ClassVector.new(0, length, 0)
  end

  -- local radius = celestialPos:subtract(lengthVec):divide(2)

  local self = ClassPath:new({
    type          = "linear",
    length        = length,
    frequency     = frequency,
    baseDirection = direction,
    offset        = celestialPos,
    direction     = (invertDirection and invertDirection == 1 and -1) or 1,
    angle = 0
    }
  )

  --Set angle to be in anti-phase to original
  if self.direction == -1 then
    self.angle = math.pi
  end

  -- print((invertDirection and invertDirection == 1 and -1) or 1)

  setmetatable(self, ClassLinearPath)
  return self
end

--Inheritance
setmetatable(ClassLinearPath, {__index = ClassPath})

function ClassLinearPath:update(dt)
  --[[
  Finds the next position this object will be in, and returns it
  ]]--
  self.angle = self.angle + (dt*self.frequency) * 2*math.pi

  if self.baseDirection == "X" then
    return ClassVector.new(self.offset:getX() + self.length*math.cos(self.angle), self.offset:getY(), self.offset:getZ())
  elseif self.baseDirection == "Y" then
    return ClassVector.new(self.offset:getX(), self.offset:getY() + self.length*math.cos(self.angle), self.offset:getZ())
  end
end

function ClassLinearPath:reset()
  --[[
  Resets the path to the beginning
  ]]--
  self.angle = self.direction == -1 and math.pi or 0
end

function ClassLinearPath:getBounds()
  if self.baseDirection == "X" then
    return {
      self.offset:getX() + self.length,
      self.offset:getX() - self.length
    }
  elseif self.baseDirection == "Y" then
    return {
      self.offset:getY() + self.length,
      self.offset:getY() - self.length
    }
  end
end

function ClassLinearPath:getDirection()
  return self.baseDirection
end

-----------------------------------------------------------------------------------------------------------------------------------------------

local function newCircularPath(celestial, parentCelestial, invertDirection, frequency)
  return ClassCircularPath:new(celestial, parentCelestial, invertDirection, frequency)
end

local function newLinearPath(celestial, invertDirection, frequency, length, direction)
  return ClassLinearPath:new(celestial, invertDirection, frequency, length, direction)
end

local function newPath(pathType, invertDirection, celestial, parentCelestial, frequency, direction, length)
  --[[
  Crates and returns a path which is not known beforehand
  ]]--
  if pathType == "circular" then
    return newCircularPath(celestial, parentCelestial, invertDirection, frequency)
  elseif pathType == "linear" then
    return newLinearPath(celestial, invertDirection, frequency, length, direction)
  end
end

return {newCircularPath = newCircularPath, newPath = newPath}
