local ClassVector = require("scenario/gravitationalRacing/classes/classVector")
local factors = require("scenario/gravitationalRacing/dataValues/factors")
local checkpointsHandler = require("scenario/gravitationalRacing/scenario/track/checkpointsHandler")

ClassVehicle = {}
ClassVehicle.__index = ClassVehicle

function ClassVehicle:new(name, position)
  if not name or not position then
    error("One or more parameters are nil")
  end

  local self = {}
  setmetatable(self, ClassVehicle)

  self.name = name
  self.mass = 1
  self.position = ClassVector.new(position.x, position.y, position.z)
  self.initialPosition = self.position:clone()
  self.acceleration = ClassVector.new(0, 0, 0)
  self.isBoosting = false

  self.pendReset = false
  self.resetTimer = 0

  self.immune = false
  self.immuneTime = 0

  return self
end

function ClassVehicle:instanceOf()
  return "ClassVehicle"
end

function ClassVehicle:getName()
  return self.name
end

function ClassVehicle:getPosition()
  return self.position
end

function ClassVehicle:getMass()
  return self.mass
end

function ClassVehicle:isResetting()
  return self.pendReset
end

function ClassVehicle:isImmune()
  return self.immune
end

function ClassVehicle:ignite()
  --[[
  Ignites the vehicle in flames
  ]]--
  scenetree.findObject("scenario_player0"):queueLuaCommand("fire.explodeVehicle()")
end

function ClassVehicle:scheduleReset()
  --[[
  Schedules the reset of the vehicle
  ]]--
  if not self.pendReset then
    local helper = require("scenario/scenariohelper")
    helper.flashUiMessage("Resetting...", 3)
    self.pendReset = true
    self.resetTimer = 2
  end
end

function ClassVehicle:applyAccel()
  --[[
  Applies wind to the vehicle
  ]]--
  --Only validate for calculated values, not specified ones
  if not self.isBoosting then
    -- Clamps a number to within a certain range
    local clamp = function(low, value, high)
      return math.min(math.max(value, low), high)
    end
    --A really high/low value will eviscerate the vehicle, and probably cause instability, so limit the wind speed
    self.acceleration:setX(clamp(-200, self.acceleration:getX(), 200))
    self.acceleration:setY(clamp(-200, self.acceleration:getY(), 200))
    self.acceleration:setZ(clamp(-200, self.acceleration:getZ(), 200))
  end

  -- print("a:"..self.acceleration:toString())

  scenetree.findObject(self.name):queueLuaCommand("obj:setWind("..self.acceleration:getX()..", "..self.acceleration:getY()..", "..self.acceleration:getZ()..")")

  if not self.isBoosting then
    --Reset acceleration
    self.acceleration:zero()
  end
end

function ClassVehicle:boost(state, ax, ay)
  --[[
  Start or stop boost the vehicle
  ]]--
  if state == nil then
    log("E", "gravitationalRacing: boost()", "Cannot change to a nil state, ignoring...")
    return
  end

  self.isBoosting = state

  if ax and ay then
    self.acceleration:setX(ax)
    self.acceleration:setY(ay)
    self.acceleration:setZ(0)
  end
end

function ClassVehicle:addWind(force)
  --[[
  Adds another amount of wind to the vehicles acceleration
  ]]--
  if not force then
    log("E", "Solar System: addWind()", "Cannot apply a nil force, ignoring...")
    return
  end

  --While boosting, no celestial affects the vehicle. This is because the acceleration
  --via a trigger ticks slower than the update time, which causes the boost to be
  --null and voided
  if not self.isBoosting then
    local a = force:multiply(50000)
    a = a:divide(factors.getDistanceScaleFactor())

    self.acceleration = self.acceleration:add(a)
  end
end

function ClassVehicle:update(dt)
  --[[
  Updates the vehicle by applying wind
  Returns true if the vehicle needs to be reset
  ]]--
  self:applyAccel()

  local objPos = scenetree.findObject(self.name):getPosition()
  self.position:setX(objPos.x)
  self.position:setY(objPos.y)
  self.position:setZ(objPos.z)

  if self.pendReset then
    if self.resetTimer <= 0 then
      --This prevents resetting happening twice as the celestial still thinks
      --the vehicle is touching it
      if self.resetTimer == -1 then
        self.pendReset = false
      else
        self.resetTimer = -1
        checkpointsHandler.resetToCheckpoint()
        self.immune = true
        self.immuneTime = 1
      end
    else
      self.resetTimer = self.resetTimer - dt
    end
  elseif self.immune then
    self.immuneTime = self.immuneTime - dt
    if self.immuneTime <= 0 then
      self.immune = false
    end
  end

  return false
end

function ClassVehicle:resetPosition()
  self.position = self.initialPosition:clone()
end

function ClassVehicle:toString()
  --[[
  Returns a string representation of this instance
  ]]--
  return "ClassVehicle[Name: "..self.name..", Mass: "..self.mass..", Position: "..self.position:toString().."]"
end

-----------------------------------------------------------------------------------------------------------------------------------------------

local function new(name, position)
  return ClassVehicle:new(name, position)
end

return {new = new}
