local M = {}

local fileHandler             = require("scenario/gravitationalRacing/utils/fileHandler")
local shortcutHandler         = require("scenario/gravitationalRacing/scenario/track/shortcutHandler")
local tableComp               = require("scenario/gravitationalRacing/utils/tableComprehension")
local scenarioDetailsHandler  = require("scenario/gravitationalRacing/scenario/scenarioDetails")

local function stringToSecs(s)
  --[[
  Converts a string to seconds
  ]]--
  local sections = {}
  local i = 1
  for part in string.gmatch(s, "%d+") do
    sections[i] = part
    i = i + 1
  end

  local totalSeconds = sections[1] * 60 + sections[2] + sections[3]/1000
  return totalSeconds
end

local function stringToMilliSecs(s)
  --[[
  Converts a string to milliseconds
  ]]--
  return stringToSecs(s) * 1000
end

local function timeToString(t)
  --[[
  Converts a time (in ms) to a string of format minutes:seconds.milliseconds
  t must be in the format of seconds.milliseconds (scenario timer standard)
  ]]--
  local format = function(value, numDigits)
    --[[
    Adds extra zeros onto the front of a number until it is n digits long
    ]]--
    local strVal = tostring(value)
    local digits = #strVal

    if digits < numDigits then
      for i = digits, numDigits-1 do
        strVal = "0"..strVal
      end
    elseif digits > numDigits then
      return math.floor(value * 10^numDigits) / 10^numDigits
    end

    return strVal
  end

  local seconds = format(math.floor(t/1000) % 60, 2)
  local minutes = format(math.floor(((t / (1000*60)) % 60)), 2)
  local milliseconds = format(math.floor(t - minutes*1000*60 - seconds*1000), 3)

  return minutes..":"..seconds.."."..milliseconds
end

local function fasterTime(t1, t2)
	--[[
	Returns the faster time of two string formatted times
	]]--
	--Check for unspecified times
	if t1 == nil then return t2 end
	if t2 == nil then return t1 end
	--Check for non-set times (if they are the same then it doesn't matter which it returns)
	if t1 == "0:00.000" then return t2 end
	if t2 == "0:00.000" then return t1 end

	local t1Secs, t2Secs = stringToSecs(t1), stringToSecs(t2)

	return t1Secs < t2Secs and t1 or t2
end

local function getUnlocksFromMedals(mapData)
  --[[
  Returns the unlocks from having x amount of medals
  ]]--
  local unlocks = {}
  local medals = fileHandler.getNumberOfMedals()

  for scName, data in pairs(mapData) do
    if not data.price or data.price <= medals then
      table.insert(unlocks, scName)
    end
  end

  return unlocks
end

local function processUnlocks(scenarioName)
  --[[
  Finds which scenarios are now unlocked and returns them formatted for the UI
  ]]--
  local mapData = jsonDecode(readFile("lua/scenario/gravitationalRacing/dataValues/trackRequirementMap.json"))
  --Unlocks from completing the scenario
  local unlocks = mapData[scenarioName].unlocks
  local medalUnlocks = getUnlocksFromMedals(mapData)

  local unlocksFormatted = {}
  for i, scName in ipairs(tableComp.mergeAppend(unlocks, medalUnlocks)) do
    local alreadyUnlocked = fileHandler.readFromFile(scName, "unlocked")

    if not alreadyUnlocked then
      local difficulty = ""
      if scName == "Solar System Simulation" then
        difficulty = jsonDecode(readFile("levels/smallgrid/scenarios/gravitationalRacing/"..scName:gsub(" ", "_"):lower()..".json"))[1].name:match("%((%a+)%)")
      else
        --Same as above
        difficulty = jsonDecode(readFile("levels/smallgrid/scenarios/gravitationalRacing/tracks/"..scName:gsub(" ", "_"):lower()..".json"))[1].name:match("%((%a+)%)")
      end

      table.insert(unlocksFormatted, {name = scName, difficulty = difficulty, colour = scenarioDetailsHandler.difficultyToColour(difficulty)})

      --Update the savefile to let it know this track has been unlocked
      fileHandler.saveToFile("scenario", {scenarioName = scName, scenarioData = {unlocked = true}})
    end
  end

  return unlocksFormatted
end

local function finish(resets, sourceFile)
  --[[
  Finishes the scenario
  ]]--
  local scenarioName, scenarioDifficulty = scenarioDetailsHandler.getScenarioDetails(sourceFile)

  local targetData = jsonDecode(readFile("lua/scenario/gravitationalRacing/dataValues/targetData.json"))[scenarioName]

  --Timer is in seconds, so convert to ms
  local time = scenario_scenarios.getScenario().timer*1000

  if scenarioName == "Tutorial" then
    fileHandler.saveToFile("scenario", {scenarioName = scenarioName, scenarioData = {tried = true, completed = true}}, false)

    --No need to store the result as the scenario will not be showing the UI
    processUnlocks(scenarioName)

    --Skip end screen and load back into hub-world
    scenario_scenariosLoader.start(scenario_scenariosLoader.loadScenario("levels/smallgrid/scenarios/gravitationalRacing/hubworld.json", nil, "levels/smallgrid/scenarios/gravitationalRacing/hubworld.json"))
    return
  end

  local targetTime = stringToMilliSecs(targetData.time)
  local bestTime = stringToMilliSecs(fileHandler.readFromFile(scenarioName, "time"))

  local targetResets = targetData.resets
  local bestResets = fileHandler.readFromFile(scenarioName, "resets")

  local collectedCollectables = shortcutHandler.hasCollectedCollectables()

  --Update savefile for this scenario
  fileHandler.saveToFile("scenario", {
    scenarioName = scenarioName,
    scenarioData = {
      tried = true,
      completed = true,
      resets = resets,
      time = timeToString(time),
      medals = {
        time = time <= targetTime,
        resets = resets <= targetResets
      },
      collectables = collectedCollectables
    }
  }, false)

  local unlocksFormatted = processUnlocks(scenarioName)

  local medalsGottenAlready = fileHandler.readFromFile(scenarioName, "medals")

  --Accumulate the data for the UI
  local scData = {
    scenarioName = scenarioName:upper(),
    difficulty = {
      type = scenarioDifficulty:upper(),
      colour = scenarioDetailsHandler.difficultyToColour(scenarioDifficulty)
    },
    times = {
      set             = time,
      best            = bestTime,
      target          = targetTime,
      targetFormatted = targetData.time
    },
    resets = {
      set    = resets,
      best   = bestResets,
      target = targetResets
    },
    unlocks = unlocksFormatted,
    medalsGottenAlready = medalsGottenAlready
  }

  scenario_scenarios.stopRaceTimer()
  bullettime.set(1/8)

  guihooks.trigger("gravitationalRacingScenarioEndScreen", scData)
end

M.fasterTime = fasterTime
M.finish = finish
return M
