local M = {}

local function createNewSaveFile()
	--[[
	Creates a blank new save file
	]]--
	local content = {}
	local fileDirectory = "/levels/smallgrid/scenarios/gravitationalRacing/tracks/"
	local files = FS:findFilesByPattern(fileDirectory, "*.json", 1, true, false)

	content.scenarios = {["Solar System Simulation"] = {unlocked = false}}
	for _, fileName in pairs(files) do
		local scenarioData = jsonReadFile(fileName)[1]
		local scenarioName = scenarioData.name:match("[%a%s]*")
		--Remove space at the end of the name
		scenarioName = scenarioName:sub(1, #scenarioName-1)

		local scenarioFilename = fileName:gsub(fileDirectory, ""):gsub(".json", "")
		if scenarioName == "Tutorial" then
			content.scenarios[scenarioName] = {
				unlocked = true,
				tried = false,
				completed = false
			}
		else
			content.scenarios[scenarioName] = {
				unlocked = false,
				tried = false,
				completed = false,
				resets = -1,
				time = "0:00.000",
				medals = {
					time = false,
					resets = false
				},
				collectables = {
					false
				}
			}
		end

	end

	serializeJsonToFile("gr_savefile", content, true)
end

local function checkFileExists()
	--[[
	Checks to see if it can open the savefile. If it can then also return the file
	so any function calling this doesn't have to open the file again - efficiency
	]]--
	local file = readFile("gr_savefile")
	if file then
		return true, file
	else
		return false
	end
end

local function readWholeFile()
	--[[
	Checks the file exists and it up to date, and then reads the file
	]]--
	local fileExists, file = checkFileExists()
	if not fileExists then
		createNewSaveFile()
		--Update the file variable with newly created info
		_, file = checkFileExists()
	end

	return jsonDecode(file, "gr_savefile")
end

local function readFromFile(scName, attribute)
	--[[
	Returns data from the save file
	If the attribute parameter is not given, this functions returns the whole scenario
	data for that scName
	Else, returns a specific attribute
	]]--
	if not scName then
		error("scName is nil")
		return
	end

	local fileContents = readWholeFile().scenarios[scName]

	if not attribute then
		return fileContents
	else
		return fileContents[attribute]
	end
end

local function getNumberOfMedals()
	--[[
	Returns the numnber of medals + collectables achieved
	]]--
	local contents = readWholeFile().scenarios
	local numMedals = 0
	for k, scData in pairs(contents) do
		--Some scenarios do not have medals
		if scData.medals then
			for _, medalAchieved in pairs(scData.medals) do
				if medalAchieved then
					numMedals = numMedals + 1
				end
			end
		end

		if scData.collectables then
			if scData.collectables.one then
				numMedals = numMedals + 1
			end
		end
	end

	return numMedals
end

local function saveToFile(type, data, forceOverwrite)
	--[[
	Saves data to the savefile
	forceOverwrite is an optional parameter and will force the update on the savefile,
	irregardless of whether it is better than previous
	]]--
	if not (type == "scenario" or type == "statistics") then
		error("This data cannot be stored to file! [type="..(type and type or "nil").."]")
	end

	local saveFileData = readWholeFile()

	if type == "scenario" then
		local scName, scData = data.scenarioName, data.scenarioData
		if not scName or not scData then
			error("Invalid save data!")
		else
			local newData = {}
			if scName == "Tutorial" then
				newData = {
					unlocked = true,
					tried = scData.tried or false,
					completed = scData.completed or false,
				}
			elseif scName == "Solar System Simulation" then
				newData = {
					unlocked = scData.unlocked or false,
				}
			else
				local currentData = saveFileData.scenarios[scName]
				--Hub world will not have current data
				if not currentData then
					return
				end

				local resets, unlckd, cmpltd, time, tried, clltbls = scData.resets, scData.unlocked, scData.completed, scData.time, scData.tried, scData.collectables
				local medals = scData.medals
				--Apply new data, even if the new data is worst than the old
				if forceOverwrite then
					newData = {
						resets     = resets or -1,
						unlocked   = unlckd or false,
						completed  = cmpltd or false,
						tried      = tried  or false,
						time       = time   or "0:00:000"
					}

					--Some sceanrios do not have collectables
					if clltbls ~= nil then
						newData.collectables = {}
						if type(clltbls) == "boolean" then
							newData.collectables[1] = clltbls
						elseif type(clltbls) == "table" then
							for i, v in ipairs(clltbls) do
								newData.collectables[i] = v
							end
						end
					end
				else
					local betterBool = function(curr, new, best)
						--[[
						Returns the best boolean value, with respect to what is considered the
						best (true or false)
						Note: if new is nil, returns curr
						]]--
						if new == nil then
							return curr
						end

						if curr == best then
							return curr
						elseif new == best then
							return new
						else
							--Both new and curr are the same
							return curr
						end
					end

					local smallestSetValue = function(curr, new)
						--[[
						Returns the smallest set value
						]]--
						if new == nil then
							return curr
						elseif curr == -1 then
							return new
						end

						if curr > new then
							return new
						elseif curr < new then
							return curr
						else
							--Both new and curr are the same
							return new
						end
					end

					newData = {
						resets    = smallestSetValue(currentData.resets, resets),
						unlocked  = betterBool(currentData.unlocked,  unlckd, true),
						completed = betterBool(currentData.completed, cmpltd, true),
						tried     = betterBool(currentData.tried,     tried,  true),
						time      = require("scenario/gravitationalRacing/scenario/scenarioEndHandler").fasterTime(currentData.time, time),
						medals = {
							time   = betterBool(currentData.medals.time,   medals and medals.time   or nil, true),
							resets = betterBool(currentData.medals.resets, medals and medals.resets or nil, true)
						}
					}

					--Not necessary anymore?
					if currentData.collectables then
						newData.collectables = {}
						if not clltbls then
							--No update necessary
							newData.collectables = currentData.collectables
						else
							for i, v in ipairs(clltbls) do
								newData.collectables[i] = betterBool(currentData.collectables[i], v, true)
							end
						end
					end
				end
			end

			--Error checking to prevent deleting data from the savefile
			if scName ~= "Tutorial" and scName ~= "Solar System Simulation" then
				local keys = {"completed", "tried", "unlocked", "time", "resets", "medals", "collectables"}
				--Make sure that every key is in the new data (does not check contents)
				for _, key in ipairs(keys) do
					local found = false
					for k, _ in pairs(newData) do
						if key == k then
							found = true
							break
						end
					end

					--Key not found
					if not found then
						log("E", "gravitationalRacing: saveToFile()", "Save data incomplete, ignoring...")
						--Maybe let the user know?
						return
					end
				end
			end

			saveFileData.scenarios[scName] = newData

			serializeJsonToFile("gr_savefile", saveFileData, true)
		end
	end
end

M.createNewSaveFile = createNewSaveFile
M.readWholeFile = readWholeFile
M.readPlayerData = readPlayerData
M.readFromFile = readFromFile
M.getNumberOfMedals = getNumberOfMedals
M.saveToFile = saveToFile
return M
