local M = {}

local function mergeAppend(a, b)
  --[[
  Merges two interge index tables into a new table, preventing overwriting
  of the same index
  ]]--
  local c = {}
  local index = 0

  for i, v in ipairs(a) do
    c[i] = v
    index = i
  end

  for i, v in ipairs(b) do
    c[index + 1] = v
    index = index + 1
  end

  return c
end

local function merge(a, b)
  --[[
  Merges table b into a. If b is nil, then it returns a
  This does not merge in place, so can be used to clone a table
  ]]--
  local c = {}

  for k, v in pairs(a or {}) do
    c[k] = v
  end

  for k, v in pairs(b or {}) do
    c[k] = v
  end

  return c
end

local function lengthOfTable(T)
  local c = 0
  for _, _ in pairs(T) do
    c = c + 1
  end

  return c
end

local function contains(T, v)
  --[[
  Returns whether a table T contains a value v
  This function is recursively defined
  ]]--
  if not v then
    error("Cannot check for a nil value!")
  end

  for _, value in ipairs(T) do
    if type(value) == "table" then
      if contains(value, v) then
        return true
      end
    else
      if value == v then
        return true
      end
    end
  end

  return false
end

local function flatten(T)
  --[[
  Flatterns a table T by removing all subtables in T whilst maintaining order
  This function is recursively defined
  ]]--
  --Base Case
  if type(T) ~= "table" then
    return {T}
  end

  --General Case
  local flattenedT = {}
  for _, element in ipairs(T) do
    for _, value in ipairs(flatten(element)) do
      --Value is now a value and not a table (if it was previously)
      table.insert(flattenedT, value)
    end
  end

  return flattenedT
end

local function containsIndex(T, v, indexes)
  --[[
  Returns the index (multiple for sub-tables) for a particular value in a table
  This function is recursively defined
  ]]--
  if not v then
    error("Cannot check for a nil value!")
  end

  if not indexes then
    indexes = {}
  end

  for i, value in ipairs(T) do
    if type(value) == "table" then
      local subindex = containsIndex(value, v, indexes)
      if subindex then
        return flatten({i, subindex})
      end
    else
      if value == v then
        return i
      end
    end
  end

  return nil
end

M.mergeAppend = mergeAppend
M.merge = merge
M.lengthOfTable = lengthOfTable
M.contains = contains
M.flatten = flatten
M.containsIndex = containsIndex
return M
